/*******************************************************************************
 * Copyright (c) 2006, 2014 Tom Schindl and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Tom Schindl <tom.schindl@bestsolution.at> - initial API and implementation
 *     Florian Potschka <signalrauschen@gmail.com> - Bug 260061
 *     Alexander Ljungberg <siker@norwinter.com> - Bug 260061
 *     Jeanderson Candido <http://jeandersonbc.github.io> - Bug 414565
 *     Alexander Mark Diewald <diewald@fortiss.org>
 *******************************************************************************/
package org.fortiss.tooling.base.ui.annotation.labelprovider;

import static org.fortiss.tooling.base.utils.SystemUtils.isMacOSXPlatform;

import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.swt.graphics.Image;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.ui.annotation.view.IAnnotationViewPart;

/**
 * <p>
 * A {@link ColumnLabelProvider} for annotations that displays checkboxes indicating the value of
 * annotations that contain boolean values.
 * </p>
 * <p>
 * This implementation uses 'screenshot' {@link Image}s of the checkboxes, since {@link TableViewer}
 * is not able to display SWT widgets when a cell is not edited. Since the screenshots are made
 * during runtime, the resulting images fit the underlying OS.
 * </p>
 * 
 * @author diewald
 */
public class CheckBoxLabelProvider extends AnnotationLabelProvider {

	/** Constructor. */
	public CheckBoxLabelProvider(Class<? extends IAnnotatedSpecification> clazz,
			IAnnotationViewPart viewPart, EditingSupport editingSupport) {
		super(clazz, viewPart, editingSupport);
	}

	/** {@inheritDoc} */
	@Override
	public String getText(Object element) {
		// See #2443
		if(isMacOSXPlatform() && element instanceof AnnotationEntry &&
				((AnnotationEntry)element).canEdit(clazz)) {
			return isChecked(element) ? "[X]" : "[ ]";
		}

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Image getImage(Object element) {
		if(element instanceof AnnotationEntry) {
			boolean enabled = ((AnnotationEntry)element).canEdit(clazz);
			return org.fortiss.tooling.base.ui.viewers.CheckBoxLabelProvider
					.getImage(isChecked(element), enabled);
		}
		return null;
	}

	/**
	 * Determines whether the boolean value is 'true' or 'false'. Used to determine the 'checked'
	 * status of the checkbox.
	 */
	protected boolean isChecked(Object element) {
		if(element instanceof AnnotationEntry) {
			AnnotationEntry annotationEntry = (AnnotationEntry)element;

			Object value = null;
			try {
				value = annotationEntry.getSpecificationValue(clazz);
			} catch(Exception e) {
				return false;
			}
			if(value != null) {
				// It is already known that the value is a boolean.
				return (Boolean)value;
			}
		}
		return false;
	}
}
