/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.dialog;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureDescriptor;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionInputParameter;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionInputParameters;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionReqInputParameter;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionReqInputParameters;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.utils.BaseMathUtils;

/**
 * {@link Dialog} that has a list of {@link Label}s and {@link Text} fields whose number matches the
 * size of the loaded {@link AnnotationActionInputParameters}. The entries of the loaded
 * {@link AnnotationActionInputParameters} define the displayed text and the expected value types.
 * 
 * For now only a subset of {@link Number}s are allowed as values.
 * 
 * @author diewald
 */
public class MultiFieldInputDialog extends Dialog {

	/** Container composite for the fields of this dialog. */
	private Composite container;

	/** List of textual descriptors and the expected types. */
	private AnnotationActionReqInputParameters parameters;

	/** {@link IAnnotatedSpecification} whose values shall be modified. */
	private IAnnotatedSpecification annotation;

	/** List of textual descriptors and the expected types. */
	private AnnotationActionInputParameters parameterValues = new AnnotationActionInputParameters();

	/** List of entered parameters. */
	private Map<Text, AnnotationActionInputParameter> fieldParamAssoc = new HashMap<>();

	/** Constructor. */
	public MultiFieldInputDialog(Shell parentShell) {
		super(parentShell);
	}

	/** {@inheritDoc} */
	@Override
	protected Control createDialogArea(Composite parent) {
		Composite area = (Composite)super.createDialogArea(parent);
		container = new Composite(area, SWT.NONE);
		container.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		GridLayout layout = new GridLayout(2, false);
		container.setLayout(layout);

		addAllInputFields(parameters);

		return area;
	}

	/**
	 * Loads the given list of input parameter descriptors whose gui elements are created after the
	 * dialog has been opened.
	 */
	public void loadAllInputFields(AnnotationActionReqInputParameters inputParameterList,
			IAnnotatedSpecification annotation) {
		parameters = inputParameterList;
		this.annotation = annotation;
	}

	/** Adds all labels and input fields described in the {@code parameterList}. */
	protected void addAllInputFields(AnnotationActionReqInputParameters inputParameterList) {
		for(AnnotationActionReqInputParameter curParameter : inputParameterList) {
			EStructuralFeatureDescriptor featureDescr = curParameter.getSecond();
			if(featureDescr != null) {
				Class<?> inputParamType = featureDescr.getEType(annotation).getInstanceClass();
				AnnotationActionInputParameter inputParam =
						new AnnotationActionInputParameter(curParameter.getFirst(), inputParamType);
				parameterValues.add(curParameter.getFirst(), inputParamType);
				addInputFields(inputParam);
			}
		}
	}

	/**
	 * Add the gui elements of an input field consisting of a textual descriptor and a Text box.
	 * 
	 * @param parameter
	 *            describes the expected input.
	 */
	protected void addInputFields(AnnotationActionInputParameter parameter) {
		Label parameterLabel = new Label(container, SWT.NONE);
		parameterLabel.setText(parameter.getFirst());

		GridData gridInputParameter = new GridData();
		gridInputParameter.grabExcessHorizontalSpace = true;
		gridInputParameter.horizontalAlignment = GridData.FILL;

		Text parameterInput = new Text(container, SWT.BORDER);
		parameterInput.setLayoutData(gridInputParameter);
		fieldParamAssoc.put(parameterInput, parameter);
	}

	/** Returns the list of parameters entered in the dialog. */
	public AnnotationActionInputParameters getParameterValues() {
		return parameterValues;
	}

	/** {@inheritDoc} */
	@Override
	protected void okPressed() {
		for(Entry<Text, AnnotationActionInputParameter> curInputField : fieldParamAssoc
				.entrySet()) {
			String enteredValue = curInputField.getKey().getText();
			@SuppressWarnings("unchecked") Class<Number> expectedType =
					(Class<Number>)curInputField.getValue().getSecond();
			Number value = BaseMathUtils.getNumberFromString(enteredValue, expectedType);

			parameterValues.put(curInputField.getValue().getFirst(), expectedType, value);
		}
		super.okPressed();
	}
}
