/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.layout.auto;

import static java.lang.Math.max;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_CONNECTOR_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodeSize;
import static org.fortiss.tooling.kernel.ui.util.KernelUIUtils.getName;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.Label;
import org.eclipse.gef.EditPartViewer;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.ui.editor.DiagramEditorBase;
import org.fortiss.tooling.base.ui.editpart.ConnectionEditPartBase;
import org.fortiss.tooling.base.ui.editpart.ElementEditPartBase;

/**
 * {@link IAutoLayouterTapeMeasure} specialization to for automatic layout of models to be displayed
 * in {@link DiagramEditorBase}-based editors.
 * 
 * @author barner
 */
public class DiagramTapeMeasure implements IAutoLayouterTapeMeasure {

	/** {@link EditPartViewer} used to display the underlying diagram. */
	private EditPartViewer diagramEpv;

	/** Underlying GC. */
	private GC gc;

	/** Constructor. */
	public DiagramTapeMeasure(EditPartViewer epv, GC gc) {
		this.diagramEpv = epv;
		this.gc = gc;
	}

	/** Returns the {@code Font} used by the given {@code element} (may be {@code null}). */
	private Font getFont(IModelElement element) {
		if(diagramEpv == null) {
			return null;
		}
		Object ep = diagramEpv.getEditPartRegistry().get(element);
		if(ep instanceof ElementEditPartBase) {
			Label label = ((ElementEditPartBase<?>)ep).createLabelFigure();
			if(label != null) {
				return label.getFont();
			}
		} else if(ep instanceof ConnectionEditPartBase) {
			IFigure figure = ((ConnectionEditPartBase<?>)ep).getFigure();
			if(figure != null) {
				return figure.getFont();
			}
		}

		return null;
	}

	/** Returns the text extend of the given {@code element}'s label (may be {@code null}). */
	private Point getTextExtent(IModelElement element) {
		if(gc == null) {
			return null;
		}

		String name = getName(element);
		if(name == null) {
			return null;
		}

		gc.setFont(getFont(element));
		Point rval = gc.textExtent(name);

		return rval;
	}

	/** {@inheritDoc} */
	@Override
	public int getElementWidth(IHierarchicElement element) {
		Dimension dimension = getNodeSize((ILayoutedModelElement)element);

		int width = max(DEFAULT_SHAPE_MINIMUM_WIDTH, dimension.getWidth());
		Point textExtend = getTextExtent(element);
		if(textExtend != null) {
			width = max(width, textExtend.x + 6 * DEFAULT_GRID_SIZE);
		}

		return width;
	}

	/** {@inheritDoc} */
	@Override
	public int getElementHeight(IHierarchicElement element) {
		Dimension dimension = getNodeSize((ILayoutedModelElement)element);
		int numElemConnectors = element.getConnectors().size();
		int numElemInputConnectors =
				pickInstanceOf(EntryConnectorBase.class, element.getConnectors()).size();
		int numConnectors = max(numElemInputConnectors, numElemConnectors - numElemInputConnectors);

		return max((1 + 2 * numConnectors) * DEFAULT_CONNECTOR_SIZE, dimension.getHeight());
	}

	/** {@inheritDoc} */
	@Override
	public int getConnectorWidth(IConnector connector) {
		return 2 * DEFAULT_CONNECTOR_SIZE;
	}

	/** {@inheritDoc} */
	@Override
	public int getConnectorHeight(IConnector connector) {
		return 2 * DEFAULT_CONNECTOR_SIZE;
	}

	/** {@inheritDoc} */
	@Override
	public int getConnectionWidth(IConnection connection) {
		Point textExtend = getTextExtent(connection);
		return textExtend != null ? textExtend.x : 0;
	}
}
