/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.annotation.valueprovider;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.fortiss.tooling.base.ToolingBaseActivator;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureDescriptor.EReferenceScope;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * Base class for {@link IAnnotationValueProvider}s that manages {@link EStructuralFeature}s
 * contained by a {@link IAnnotatedSpecification}.
 * 
 * @author barner
 */
public abstract class EStructuralFeatureValueProviderBase<T extends IAnnotatedSpecification>
		extends ValueProviderBase<T> {

	/**
	 * {@link EStructuralFeatureDescriptor} of the {@link IAnnotatedSpecification} managed by this
	 * {@link IAnnotationValueProvider}.
	 */
	private final EStructuralFeatureDescriptor structuralFeatureDescriptor;

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} whose
	 * storage is implemented by a single {@link EAttribute}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EAttribute eAttribute) {

		super(annotatedSpecificationEClass);

		structuralFeatureDescriptor = new EStructuralFeatureDescriptor(eAttribute);
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for an {@link IAnnotatedSpecification}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EReference eReference, EStructuralFeatureDescriptor.EReferenceScope eReferenceScope) {

		super(annotatedSpecificationEClass);

		structuralFeatureDescriptor = new EStructuralFeatureDescriptor(eReference, eReferenceScope);
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} whose
	 * storage is implemented by an {@link EClass} that contains only a single
	 * {@link EStructuralFeature}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass) {

		this(annotatedSpecificationEClass, EStructuralFeatureDescriptor.EReferenceScope.MODEL);
	}

	/**
	 * Generalizes the interface of {@link #EStructuralFeatureValueProviderBase(EClass)}.
	 * 
	 * @param annotatedSpecificationEClass
	 *            {@link EClass} for which {@link EStructuralFeatureValueProviderBase} should be
	 *            constructed.
	 * 
	 * @param eReferenceScope
	 *            {@link EReferenceScope} to be offered to the user to select values for
	 *            {@link EReference}s.
	 * 
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EStructuralFeatureDescriptor.EReferenceScope eReferenceScope) {

		super(annotatedSpecificationEClass);

		final EList<EStructuralFeature> featureList =
				annotatedSpecificationEClass.getEStructuralFeatures();
		if(featureList.size() > 1) {
			throw new RuntimeException("Annotation class " + annotatedSpecificationEClass +
					" must not contain more than one structural feature.");
		}

		if(featureList.size() == 1) {
			// There is exactly one structural feature
			EStructuralFeature structuralFeature = featureList.get(0);
			if(structuralFeature instanceof EAttribute) {
				structuralFeatureDescriptor =
						new EStructuralFeatureDescriptor((EAttribute)structuralFeature);
			} else if(structuralFeature instanceof EReference) {
				structuralFeatureDescriptor = new EStructuralFeatureDescriptor(
						(EReference)structuralFeature, eReferenceScope);
			} else {
				throw new RuntimeException(
						"Unsupported EStructuralFeature specialization: " + structuralFeature);
			}
		} else {
			structuralFeatureDescriptor = null;
		}
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} based on
	 * a custom {@link EStructuralFeatureDescriptor}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EStructuralFeatureDescriptor structuralFeatureDescriptor) {
		super(annotatedSpecificationEClass);
		this.structuralFeatureDescriptor = structuralFeatureDescriptor;
	}

	/** {@inheritDoc} */
	@Override
	public EStructuralFeatureDescriptor getEStructuralFeatureDescriptor() {
		return structuralFeatureDescriptor;
	}

	/** Returns the {@link EFactory} for the given {@link EStructuralFeature}. */
	private EFactory getEFactory(EStructuralFeature structuralFeature) {
		return structuralFeature.getEType().getEPackage().getEFactoryInstance();
	}

	/**
	 * Sets a value for a {@link IAnnotatedSpecification} from a {@link String} representation of
	 * the input.
	 */
	protected <V> void setAnnotationValueFromString(String value, T specification)
			throws Exception {

		EStructuralFeature structuralFeature =
				getEStructuralFeatureDescriptor().getEStructuralFeature(specification);

		if(structuralFeature instanceof EAttribute) {
			EFactory eFactory = getEFactory(structuralFeature);
			EDataType eAttributeType = ((EAttribute)structuralFeature).getEAttributeType();
			try {
				specification.eSet(structuralFeature,
						eFactory.createFromString(eAttributeType, value));
			} catch(Exception e) {
				// Keep previous value
			}
		} else {
			throw new Exception(
					"setAnnotationValueFromString(String, T) is only availabe for EAttributes. Use setAnnotationValue(U value, T specification) instead.");
		}
	}

	/**
	 * Sets a value for a {@link IAnnotatedSpecification} from a {@code Collection<String>}
	 * representation of the input.
	 */
	private void setAnnotationValueFromString(Collection<String> collection, T specification)
			throws Exception {
		EStructuralFeature structuralFeature =
				getEStructuralFeatureDescriptor().getEStructuralFeature(specification);

		if(structuralFeature instanceof EAttribute) {
			EFactory eFactory = getEFactory(structuralFeature);
			EDataType eAttributeType = ((EAttribute)structuralFeature).getEAttributeType();
			EList<Object> values = new BasicEList<>();
			for(String s : collection) {
				Object o = null;
				try {
					o = eFactory.createFromString(eAttributeType, s);
				} catch(Exception e) {
					error(ToolingBaseActivator.getDefault(), "Could not convert " +
							(s != null ? s : "<null>") + " to datatype " + eAttributeType + ".", e);
				}
				if(o != null) {
					values.add(o);
				}
			}
			specification.eSet(structuralFeature, values);
		} else {
			throw new Exception(
					"setAnnotationValueFromString(Collection<String>, T) is only availabe for EAttributes.");
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean canEdit(T specification) {
		return true;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> void setAnnotationValue(U value, T specification) throws Exception {
		final EStructuralFeature eStructuralFeature =
				getEStructuralFeatureDescriptor().getEStructuralFeature(specification);
		if(value instanceof String) {
			// In case the input choice is presented in dynamically managed combo box (i.e., where
			// new value choices can be entered by the user), the empty String indicates that the
			// current value should be deleted, and a neighboring value from the set should be used.
			// In case the last occurrence of a value choice has been removed, it will no longer
			// be offered in the combo box.
			if(((String)value).isEmpty()) {
				String rval = determineCurrentAnnotationValueAndUpdateInputChoice((String)value,
						specification);
				if(rval != null) {
					value = (U)rval;
				}
			}

			setAnnotationValueFromString((String)value, specification);
		} else if(value instanceof Collection<?>) {
			Collection<?> collection = (Collection<?>)value;
			if(collection.isEmpty()) {
				specification.eSet(eStructuralFeature, emptyList());
			} else {
				if(collection.iterator().next() instanceof String) {
					setAnnotationValueFromString((Collection<String>)value, specification);
				} else {
					specification.eSet(eStructuralFeature, value);
				}
			}
		} else {
			if(value != null) {
				specification.eSet(eStructuralFeature, value);
			} else {
				// Treat 'null' as special value used to unset EStructural features. This
				// implementation choice prevents to distinguish between 'null' and unset for
				// EReferences).
				specification.eUnset(eStructuralFeature);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification) throws Exception {
		// See comment in setAnnotationValue(U value, T specification)
		if((value == null || value.isEmpty())) {
			String rval = determineCurrentAnnotationValueAndUpdateInputChoice(value, specification);
			if(rval != null) {
				value = rval;
			}
		}

		EStructuralFeature structuralFeature =
				getEStructuralFeatureDescriptor().getEStructuralFeature(specification);

		if(structuralFeature instanceof EAttribute) {
			setAnnotationValueFromString(value, specification);
		} else {
			throw new Exception(
					"setAnnotationValue(String, T) is not supported / has not been implemented for the annotation type " +
							structuralFeature.getEType().getName() + "(specification class: " +
							specification.getClass() + ").");
		}
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> U getAnnotationValue(T specification) throws Exception {
		return (U)specification
				.eGet(getEStructuralFeatureDescriptor().getEStructuralFeature(specification));
	}

	/**
	 * Creates a new instance of the {@link EStructuralFeature} (used to populate newly created
	 * annotation instances created in
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}).
	 */
	protected Object createStructuralFeatureInstance() throws Exception {
		EStructuralFeature structuralFeature =
				getEStructuralFeatureDescriptor().getEStructuralFeature(null);
		if(structuralFeature == null) {
			return null;
		}

		if(structuralFeature instanceof EAttribute) {
			EFactory factory = getEFactory(structuralFeature);
			EDataType attributeEDataType = ((EAttribute)structuralFeature).getEAttributeType();

			String defaultValueLiteral = structuralFeature.getDefaultValueLiteral();
			Object defaultValue = null;
			if(defaultValueLiteral != null) {
				// Use default value if it has been specified
				defaultValue = factory.createFromString(attributeEDataType, defaultValueLiteral);
			} else {
				// Return data types default value
				defaultValue = attributeEDataType.getDefaultValue();
			}

			if(structuralFeature.getUpperBound() == 1) {
				return defaultValue;
			}

			List<Object> valueList = new ArrayList<Object>();
			for(int i = 0; i < structuralFeature.getLowerBound(); ++i) {
				valueList.add(defaultValue);
			}

			return valueList;
		} else if(structuralFeature instanceof EReference) {
			return null;
		} else {
			throw new Exception(
					"createStructuralFeatureInstance() is not supported / has not been implemented for the annotation type " +
							structuralFeature.getName() + ".");
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void decorateAnnotationSpecification(T specification) throws Exception {

		// Create and set structural feature implementing the annotation
		Object structuralFeatureVal = createStructuralFeatureInstance();
		if(structuralFeatureVal != null) {
			EStructuralFeature eStructuralFeature =
					getEStructuralFeatureDescriptor().getEStructuralFeature(specification);
			specification.eSet(eStructuralFeature, structuralFeatureVal);
		}
	}
}
