/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.extension.base;

import static org.fortiss.tooling.kernel.ui.util.ConstraintsUIUtils.triggerMarkersRefresh;
import static org.fortiss.tooling.kernel.utils.ConstraintsUtils.createOutdatedStatus;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Shell;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.ErrorConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.FailedConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedConstraintInstanceStatus;
import org.fortiss.tooling.kernel.service.IConstraintService;
import org.fortiss.tooling.kernel.ui.extension.IConstraintUI;
import org.fortiss.tooling.kernel.ui.service.IConstraintUIService;
import org.fortiss.tooling.kernel.ui.service.IConstraintUIService.IFix;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;

/**
 * Base classes for constraint verification GUI.
 * 
 * @author aravantinos
 */
public class ConstraintUIBases {

	/** Base class for constraint verification GUI. */
	public static abstract class ConstraintUIBase implements IConstraintUI {
		/** {@inheritDoc} */
		@Override
		public boolean canOpen(ConstraintInstance ci, IConstraintInstanceStatus status) {
			return status instanceof OutdatedConstraintInstanceStatus;
		}

		/** {@inheritDoc} */
		@Override
		public String getDescription(ConstraintInstance ci) {
			return this.getDescription();
		}

		/** {@inheritDoc} */
		@Override
		public List<IFix> fixes(ConstraintInstance ci, IConstraintInstanceStatus status) {
			// No fix by default, it is however strongly recommended to override in order to improve
			// user experience.
			return new ArrayList<IFix>();
		}

		/** {@inheritDoc} */
		@Override
		public boolean openStatus(ConstraintInstance ci, IConstraintInstanceStatus status) {
			if(!canOpen(ci, status)) {
				return false;
			}
			if(status instanceof OutdatedConstraintInstanceStatus) {
				IConstraintService.getInstance().verify(ci);
				IConstraintInstanceStatus currentStatus =
						IConstraintUIService.getInstance().getStatus(ci);
				if(currentStatus instanceof ErrorConstraintInstanceStatus && canOpen(ci, status)) {
					openStatus(ci, currentStatus);
				}
				return true;
			}
			if(status instanceof FailedConstraintInstanceStatus ||
					status instanceof ErrorConstraintInstanceStatus) {
				List<IFix> fixes = IConstraintUIService.getInstance().fixes(ci);
				Shell shell = Display.getCurrent().getActiveShell();
				int result = new FixDialog(shell, status, getMessage(ci, status), fixes).open();
				if(fixes.size() == 1 && result == 0) {
					// get(0) because the size is equal to 1 by the previous test
					fixes.get(0).runFix(status);
				} else if(result > 1) {
					// In this case, result = 0 iff the dialog was cancelled, 1 iff the dialog was
					// OK, or any number greater than 1 to select a given fix, hence the result-2
					// below.
					fixes.get(result - 2).runFix(status);
				}
				return true;
			}
			return false;
		}

		/** {@inheritDoc} */
		@Override
		public boolean isTimeConsuming() {
			return false;
		}

		/** Dialog presenting the error/failure and the possible fix(es). */
		private static class FixDialog extends MessageDialog {
			/** List of possible fixes. */
			List<IFix> fixes;

			/** Constructor. */
			public FixDialog(Shell parentShell, IConstraintInstanceStatus s, String msg,
					List<IFix> fixes) {
				super(parentShell, getTitle(s), null, msg, getIconType(s), getButtons(fixes), 1);
				this.fixes = fixes;
			}

			/**
			 * Returns the buttons to display: "OK" is always displayed. In addition, if there is
			 * only one fix, we provide one button for this fix. If there is more than one
			 * possible fixes, they will displayed instead horizontally (via getCustomArea).
			 */
			private static String[] getButtons(List<IFix> fixes) {
				String[] res = new String[fixes.size() == 1 ? 2 : 1];
				if(fixes.size() == 1) {
					// get(0) because the size is equal to 1 by the previous test
					res[0] = "Fix: " + fixes.get(0).getDescription();
				}
				res[res.length - 1] = "OK";
				return res;
			}

			/** Returns the icon type corresponding to <code>s</code>. */
			private static String getTitle(IConstraintInstanceStatus s) {
				if(s instanceof FailedConstraintInstanceStatus) {
					return "Verification of constraint failed";
				}
				return "Error during constraint verification";
			}

			/** Returns the icon type corresponding to <code>s</code>. */
			private static int getIconType(IConstraintInstanceStatus s) {
				return s instanceof FailedConstraintInstanceStatus ? MessageDialog.WARNING : ERROR;
			}

			/** {@inheritDoc} */
			@Override
			protected Control createCustomArea(Composite parent) {
				if(fixes.size() <= 1) {
					return null;
				}
				Group buttonGroup = new Group(parent, SWT.NONE);
				buttonGroup.setLayout(new GridLayout());
				buttonGroup.setLayoutData(new GridData(SWT.RIGHT, SWT.TOP, true, false));
				buttonGroup.setText("Possible fixes");
				for(IFix fix : fixes) {
					Button button = new Button(buttonGroup, SWT.None);
					button.setText(fix.getDescription());
					button.addSelectionListener(new SelectionAdapter() {
						@Override
						public void widgetSelected(SelectionEvent event) {
							buttonPressed(fixes.indexOf(fix) + 1);
						}
					});
				}
				// Does not matter what we return as long as it's not null
				return buttonGroup;
			}
		}

		/** {@inheritDoc} */
		@Override
		public String getMessage(ConstraintInstance ci, IConstraintInstanceStatus status) {
			IConstraintInstanceStatus s = IConstraintUIService.getInstance().getStatus(ci);
			// Default message. Inheriting classes SHOULD specialize.
			String name = IModelElementHandlerService.getInstance().getName(s.getConstraint());
			String pre = s instanceof ErrorConstraintInstanceStatus ? "Error while checking"
					: "Unsatisfied";
			return pre + " constraint" + (name == null ? "." : ": \"" + name + "\"");
		}

		/** {@inheritDoc} */
		@Override
		public void onOutdate(ConstraintInstance ci) {
			// By default, we just set the status to outdated
			createOutdatedStatus(ci);
		}

		/** {@inheritDoc} */
		@Override
		public boolean shouldBeManuallyActivated() {
			return false;
		}

		/** {@inheritDoc} */
		@Override
		public boolean displayAsWarning() {
			// By default, constraints failures are displayed as error not as warnings
			return false;
		}

		/** {@inheritDoc} */
		@Override
		public ImageDescriptor getIconImageDescriptor() {
			return null;
		}
	}

	/**
	 * Base class for constraint verifier UIs which automatically trigger a new check when the
	 * constraint gets outdated.
	 */
	public static abstract class ConstraintUIBaseAutocheck extends ConstraintUIBase {

		/** {@inheritDoc} */
		@Override
		public void onOutdate(ConstraintInstance ci) {
			Display display = Display.getDefault();
			String name = "Verifying " + ConstraintUIBaseAutocheck.this.getDescription();
			Job latestJob = new Job(name) {
				/** {@inheritDoc} */
				@Override
				protected IStatus run(IProgressMonitor monitor) {
					try {
						display.asyncExec(() -> IConstraintService.getInstance().cancel(ci));
						IConstraintService.getInstance().verify(ci);
						display.asyncExec(() -> triggerMarkersRefresh(ci));
						return Status.OK_STATUS;
					} catch(IllegalArgumentException e) {
						// This sort of exception might happen, not sure why, but we *in no case*
						// want to throw an error message to the user in case of some failure
						// resulting of a constraint check which he did not explicitly triggered
						return Status.CANCEL_STATUS;
					}
				}
			};
			// If the constraint is considered time consuming, we wait 1s before starting the job
			// because user modifications can go faster than the check.
			latestJob.schedule(isTimeConsuming() ? 1000 : 0);
		}

		/** {@inheritDoc} */
		@Override
		public boolean canOpen(ConstraintInstance ci, IConstraintInstanceStatus status) {
			return false;
		}
	}

	/** Handler for constraints. */
	public static class ConstraintHandler extends ModelElementHandlerBase<ConstraintInstance> {

		/** {@inheritDoc} */
		@Override
		public String getName(ConstraintInstance element) {
			return getDescription(element);
		}

		/** {@inheritDoc} */
		@Override
		public ImageDescriptor getIconImageDescriptor() {
			// This method will actually never be called
			return null;
		}

		/** {@inheritDoc} */
		@Override
		public ImageDescriptor getIconImageDescriptor(ConstraintInstance element) {
			return IConstraintUIService.getInstance().getIconImageDescriptor(element);
		}
	}
}
