/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.service.types;

import java.util.Comparator;
import java.util.TreeMap;

import org.eclipse.emf.ecore.EClass;

/**
 * Base class to map an EMF type to a target type. For a given {@link EClass} {@code Foo}, the
 * map considers the generated interface class {@code Foo} and the corresponding implementation
 * class {@code FooImpl} identical.
 */
public class EMFTypeMap<S, T> extends TreeMap<Class<? extends S>, T> {

	/**
	 * Suffix used by EMF to denote implementation classes of the Java interfaces corresponding
	 * to EClasses.
	 */
	private static final String IMPL_CLASS_SUFFIX = "Impl";

	/** Constructor. */
	public EMFTypeMap() {
		super(new Comparator<Class<? extends S>>() {
			@SuppressWarnings("unchecked")
			private Class<? extends S> getInterfaceType(Class<? extends S> c) {
				final String className = c.getSimpleName();
				if(className.endsWith(IMPL_CLASS_SUFFIX)) {
					final String ifaceName =
							className.substring(0, className.length() - IMPL_CLASS_SUFFIX.length());
					for(Class<?> iFace : c.getInterfaces()) {
						if(iFace.getSimpleName().equals(ifaceName)) {
							return (Class<? extends S>)iFace;
						}
					}
				}
				return c;
			}

			/** {@inheritDoc} */
			@Override
			public int compare(Class<? extends S> c1, Class<? extends S> c2) {
				Class<? extends S> i1 = getInterfaceType(c1);
				Class<? extends S> i2 = getInterfaceType(c2);

				if(i1 == i2) {
					return 0;
				}
				return i1.getCanonicalName().compareTo(i2.getCanonicalName());
			}
		});
	}
}
